/*
 * puzzle.h
 *
 *
 * Part of TREE-PUZZLE 5.2 (July 2004)
 *
 * (c) 2003-2004 by Heiko A. Schmidt, Korbinian Strimmer, and Arndt von Haeseler
 * (c) 1999-2003 by Heiko A. Schmidt, Korbinian Strimmer,
 *                  M. Vingron, and Arndt von Haeseler
 * (c) 1995-1999 by Korbinian Strimmer and Arndt von Haeseler
 *
 * All parts of the source except where indicated are distributed under
 * the GNU public licence.  See http://www.opensource.org for details.
 *
*
 */



#ifndef PACKAGE
#  define PACKAGE    "tree-puzzle"
#endif

#ifndef VERSION
#  define VERSION    "5.2-generic"
#endif

#define DATE       "July 2004"



#define STDOUT stdout

/* time check interval in seconds for ML/puzzling step status (900 = 15min) */
#  define TIMECHECK_INTERVAL 900

/* filenames */
#  define FILENAMELENGTH 2048

/* maximum exp difference, such that 1.0+exp(-TP_MAX_EXP_DIFF) == 1.0 */
#  define TP_MAX_EXP_DIFF 40

#  define INFILEDEFAULT     "infile"
#  define OUTFILEDEFAULT    "outfile"
#  define TREEFILEDEFAULT   "outtree"
#  define INTREEDEFAULT     "intree"
#  define DISTANCESDEFAULT  "outdist"
#  define TRIANGLEDEFAULT   "outlm.eps"
#  define UNRESOLVEDDEFAULT "outqlist"
#  define ALLQUARTDEFAULT   "outallquart"
#  define ALLQUARTLHDEFAULT "outallquartlh"
#  define SITELHDEFAULT     "outsitelh"
#  define SITELHBDEFAULT    "outsitelhb"
#  define SITERATEDEFAULT   "outsiterate"
#  define SITERATEBDEFAULT  "outsiterateb"
#  define OUTPARAMDEFAULT   "outparam"
#  define SUBSETDEFAULT     "insubsetmatr"
#  define OUTPTLISTDEFAULT  "outpstep"
#  define OUTPTORDERDEFAULT "outptorder"

/*
#  define INFILE     infilename
#  define OUTFILE    outfilename
#  define TREEFILE   outtreename
#   define INTREE     intreename
#  define DISTANCES  outdistname
#  define TRIANGLE   outlmname
#  define UNRESOLVED outqlistname
#  define ALLQUART   outallquartname
#  define ALLQUARTLH outallquartlhname
#  define SITELH     outsitelhname
#  define SITELHB    outsitelhbname
#  define SITERATE   outsiteratename
#  define SITERATEB  outsiteratebname
#  define OUTPARAM   outparamname
#  define SUBSET     insubsetmatrname
#  define OUTPTLIST  outpstepname
#  define OUTPTORDER outptordername
#  define FILEPREFIX fileprefixname
*/


 char INFILE        [FILENAMELENGTH];
 char OUTFILE       [FILENAMELENGTH];
 char TREEFILE       [FILENAMELENGTH];
 char TEMPFILE       [FILENAMELENGTH];
 char INTREE[FILENAMELENGTH];
 char DISTANCES       [FILENAMELENGTH];
 char TRIANGLE         [FILENAMELENGTH];
 char UNRESOLVED      [FILENAMELENGTH];
 char ALLQUART   [FILENAMELENGTH];
 char ALLQUARTLH [FILENAMELENGTH];
 char SITELH     [FILENAMELENGTH];
 char SITELHB    [FILENAMELENGTH];
 char SITERATE   [FILENAMELENGTH];
 char SITERATEB  [FILENAMELENGTH];
 char OUTPARAM      [FILENAMELENGTH];
 char SUBSET  [FILENAMELENGTH];
 char OUTPTLIST      [FILENAMELENGTH];
 char OUTPTORDER    [FILENAMELENGTH];

 char FILEPREFIX     [FILENAMELENGTH];


#define OUTFILEEXT    "puzzle"
#define TREEFILEEXT   "tree"
#define TEMPFILEEXT   "temp"
#define DISTANCESEXT  "dist"
#define TRIANGLEEXT   "eps"
#define UNRESOLVEDEXT "qlist"
#define ALLQUARTEXT   "allquart"
#define ALLQUARTLHEXT "allquartlh"
#define SITELHEXT     "sitelh"
#define SITELHBEXT    "sitelhb"
#define SITERATEEXT   "siterate"
#define SITERATEBEXT  "siterateb"
#define OUTPARAMEXT   "param"
#define SUBSETEXT     "subsetmatr"
#define OUTPTLISTEXT  "pstep"
#define OUTPTORDEREXT "ptorder"


/* resetprefix values  (xxx) */
#define RESET_NONE   0
#define RESET_INFILE 1
#define RESET_INTREE 2

/* setprefix_optn values  (xxx) */
#define SETPREFIX_NONE   0
#define SETPREFIX_INFILE 1
#define SETPREFIX_INTREE 2
#define SETPREFIX_RESET  3
#define SETPREFIX_PREFIX 4

/* auto_aamodel/auto_datatype values  (xxx) */
#define AUTO_OFF      0
#define AUTO_GUESS    1
#define AUTO_DEFAULT  2


/* qptlist values  (xxx) */
#define PSTOUT_NONE      0
#define PSTOUT_ORDER     1
#define PSTOUT_LISTORDER 2
#define PSTOUT_LIST      3

/* dtat_optn values  (xxx) */
#define NUCLEOTIDE 0
#define AMINOACID  1
#define BINARY     2

/* typ_optn values  (xxx) */
#define TREERECON_OPTN 0
#define LIKMAPING_OPTN 1
#define NUMTYPES 2

/* puzzlemodes (xxx) */
#define QUARTPUZ 0
#define USERTREE 1
#define CONSENSUS 2
#define PAIRDIST 3
#define NUMPUZZLEMODES 4

/* modes for rootsearch (xxx) */
#define ROOT_USER 0
#define ROOT_AUTO 1
#define ROOT_DISPLAYED 2
#define NUMROOTSEARCH 3


/* defines for types of quartet likelihood computation (xxx) */
#define EXACT  0
#define APPROX 1

/* defines for results of sequence number check (seqnumcheck) */
#define SEQNUM_OK 0
#define SEQNUM_TOOFEW 1
#define SEQNUM_TOOMANY 2



/* variables */
 cmatrix Seqchars;     /* characters contained in data set */
 ivector Seqgapchar;   /* counter for gaps contained in sequence */
 ivector Seqotherchar; /* counter for ambiguous contained in sequence */
 cmatrix treepict;     /* picture of consensus tree */
 double minscore;      /* value of edgescore on minedge */
 double tstvf84;       /* F84 transition/transversion ratio */
 double tstvratio;     /* expected transition/transversion ratio */
 double yrtsratio;     /* expected pyrimidine/purine transition ratio */
 dvector ulkl;         /* log L of user trees */
 dmatrix allsites;     /* log L per sites of utrees (numutrees,Numptrn) */
 dvector ulklc;        /* log L of utrees (clock) (numutrees,Numptrn) */
 dmatrix allsitesc;    /* log L per sites of user trees (clock) */
 FILE *utfp;           /* pointer to user tree file */
 FILE *ofp;            /* pointer to output file */
 FILE *seqfp;          /* pointer to sequence input file */
 FILE *tfp;            /* pointer to tree file */
 FILE *dfp;            /* pointer to distance file */
 FILE *trifp;          /* pointer to triangle file */
 FILE *unresfp;        /* pointer to file with unresolved quartets */
 FILE *tmpfp;          /* pointer to temporary file */
 FILE *qptlist;        /* pointer to file with puzzling step trees */
 FILE *qptorder;       /* pointer to file with unique puzzling step trees */
 int SHcodon;          /* whether SH should be applied to 1st, 2nd codon positions */
 int utree_optn;       /* use first user tree for estimation */
 int listqptrees;      /* list puzzling step trees */
 int approxqp;         /* approximate QP quartets */

 int codon_optn;       /* declares what positions in a codon should be used */
 int compclock;        /* computation of clocklike branch lengths */
 int chooseA;          /* leaf variable */
 int chooseB;          /* leaf variable */
 int clustA, clustB, clustC, clustD; /* number of members of LM clusters */

 int Frequ_optn;       /* use empirical base frequencies */

/* PSTEP_ORIG_H */
 int Maxbrnch;         /* 2*Maxspc - 3 */

 int Maxseqc;          /* number of sequence characters per taxum */
 int mflag;            /* flag used for correct printing of runtime messages */

 int numclust;         /* number of clusters in LM analysis */
 int outgroup;         /* outgroup */
 int puzzlemode;       /* computation of QP tree and/or ML distances */
 int rootsearch;       /* how location of root is found */
 int rhetmode;         /* model of rate heterogeneity */
 int seqnumcheck;      /* result of sequence number checkheterogeneity  */
 int usebestq_optn;    /* use only best quartet topology, no bayesian weights */
 int printrmatr_optn;  /* print rate matrix to screen */
 int usebranch_optn;   /* use branch lengths given in usertree */
 int show_optn;        /* show unresolved quartets    */
 int savequart_optn;   /* save memory block which quartets to file */
 int savequartlh_optn; /* save quartet likelihoods to file */
 int saveqlhbin_optn;  /* save quartet likelihoods binary */
 int savesitelh_optn;  /* save site likelihoods (PHILIP-like) */
 int savesitelhb_optn; /* save site likelihoods (binary) */
 int savesiterate_optn;  /* save site rates (PHILIP-like) */
 int savesiterateb_optn; /* save site rates (binary) */
 int readquart_optn;   /* read memory block which quartets from file */
 int fixedorder_optn;  /* use fixed puzzling insert order 1,2,3,...n */
 int conssub50_optn;   /* do  consensus down to percentage of first incongruence/ambiguity */
 int qsupport_optn;    /* compute quartet support for the splits */
 int dotreetest_optn;  /* compute tree statistics (ELW, KH, SH) */
 int dotreelh_optn;    /* compute lh/branch lengths (QP,consensus) */
 int setprefix_optn;   /* use other FILEPREFIX that INFILENAME */
 int consensus_optn;   /* do consensus tree instead of user trees */
 int sym_optn;         /* symmetrize doublet frequencies */
 int ytaxcounter;      /* counter for establishing y-coordinates of all taxa */
 int numutrees;        /* number of users trees in input tree file */
 ivector clusterA, clusterB, clusterC, clusterD;  /* clusters for LM analysis */

 ivector ycortax;      /* y-coordinates of all taxa */

 uli badqs;            /* number of bad quartets */
 uli unresqs;          /* number of unresolved quartets, formerly badqs */
 uli partresqs;        /* number of partly resolved quartets */
 uli fullresqs;        /* number of fully resolved quartets */
 uli missingqs;        /* number of fully resolved quartets */
 ulivector badtaxon;   /* involment of each taxon in a bad quartet */
 ulimatrix qinfomatr;  /* sums of quartet topologies per taxon [0]=sum */
 uli Currtrial;        /* counter for puzzling steps */

 uli Numquartets;      /* number of quartets */
 uli Numtrial;         /* number of puzzling steps */
 uli lmqts;            /* quartets investigated in LM analysis (0 = ALL) */

 int auto_datatype;       /* guess datatype ? */
 int guessdata_optn;      /* guessed datatype */

 int auto_aamodel;        /* guess amino acid modell ? */
 int guessauto_aamodel;   /* guessed amino acid modell ? */
 int guessDayhf_optn;     /* guessed Dayhoff model option */
 int guessJtt_optn;       /* guessed JTT model option */
 int guessblosum62_optn;  /* guessed BLOSUM 62 model option */
 int guessmtrev_optn;     /* guessed mtREV model option */
 int guesscprev_optn;     /* guessed cpREV model option */
 int guessvtmv_optn;      /* guessed VT model option */
 int guesswag_optn;       /* guessed WAG model option */

/* missing data: for handling subsets */
 int     readsubset_optn;           /* guessed WAG model option */
 int     Maxsubset;                 /* number of subsets */
 imatrix ss_setovlgraph;            /* size of overlap >= 3 between 2 subsets */
 imatrix ss_setoverlaps;            /* size of overlap between 2 subsets */
 imatrix ss_setovllist;             /* list with ovlerlapping subsets */
 ivector ss_setovllistsize;         /* size of list with ovlerlapping subsets */
 imatrix ss_matrix;                 /* boolean list: taxon in set? */
 imatrix ss_list;                   /* list of taxa in set */
 ivector ss_listsize;                /* size of list with taxa */

/* counter variables needed in likelihood mapping analysis */
 uli ar1, ar2, ar3;
 uli reg1, reg2, reg3, reg4, reg5, reg6, reg7;
 uli reg1l, reg1r, reg2u, reg2d, reg3u, reg3d,
 reg4u, reg4d, reg5l, reg5r, reg6u, reg6d;
 unsigned char *quartetinfo; /* place where quartets are stored */
 dvector qweight; /* for use in QP and LM analysis */
 dvector sqdiff;
 ivector qworder;
 ivector sqorder;

 int randseed;
 int psteptreestrlen;



/* prototypes */
void makeF84model(void);
void compnumqts(void);
void setoptions(void);
int  openfiletoread(FILE **, char[], char[]);
int  openfiletowrite(FILE **, char[], char[]);
int  openfiletoappend(FILE **, char[], char[]);
void closefile(FILE *);
void symdoublets(void);
void computeexpectations(void);
void putdistance(FILE *);
void findidenticals(FILE *);
double averagedist(int maxspc, dmatrix  distanmat, double *meandist, double *mindist, double *maxdist, double *stddevdist, double *vardist);
void initps(FILE *);
void plotlmpoint(FILE *, double, double);
void finishps(FILE *);
void makelmpoint(FILE *, double, double, double);
void timestamp(FILE *);
void writeoutputfile(FILE *, int);

/* definitions for writing output */
#define WRITEALL    0
#define WRITEPARAMS 1
#define WRITEREST   2

void writetimesstat(FILE *ofp);
void writecutree(FILE *, int);
void starttimer(void);
void checktimer(uli);

/* void estimateparametersnotree(void); * moved to mlparam.c */
/* void estimateparameterstree(void);   * moved to mlparam.c */

int main(int, char *[]);
int ulicmp(const void *, const void *);
int intcmp(const void *, const void *);

void readid(FILE *, int);
char readnextcharacter(FILE *, int, int);

/* skip rest of the line */
void skiprestofline(FILE *ifp,    /* input file stream */
                    int   notu,   /* taxon number      - for error msg. */
                    int   nsite); /* sequence position - for error msg. */

/* skip control characters and blanks */
void skipcntrl(FILE *ifp,    /* input file stream */
               int   notu,   /* taxon number      - for error msg. */
               int   nsite); /* sequence position - for error msg. */

/* read sequences of one data set */
void getseqs(FILE    *ifp,      /* in:  input file stream */
             int      Maxspc,   /* in:  number of taxa */
             int      Maxseqc,  /* in:  number of sites */
             cmatrix *seqch,    /* out: alignment matrix */
             cmatrix  identif); /* io:  taxon names (10 char w/o stop) */

/* initialize identifer arrays */
void initid(int      t,        /* in:  number of taxa */
            cmatrix *identif,  /* out: name array w/o end of string */
            cmatrix *namestr); /* out: name array with end of string */

/* copy undelimited identifer array to '\0' delimited identifer array */
void identif2namestr(int     num,      /* number of taxa         */
                     cmatrix Identif,  /* non-delimited names    */
                     cmatrix Namestr); /* proper delimited names */

void fputid10(FILE *, int);
int fputid(FILE *, int);

/* read first line of sequence data set */
void getsizesites(FILE *ifp,      /* in: input file stream */
                  int  *Maxspc,   /* out: number of taxa */
                  int  *Maxseqc); /* out: number of sites */

/* read alignment from file */
void readsequencefile(FILE    *seqfp,     /* in:  sequence input file stream */
                      int     *Maxspc,    /* out: number of taxa             */
                      int     *Maxseqc,   /* out: number of sites            */
                      cmatrix *identif,   /* out: name array w/o end of str  */
                      cmatrix *namestr,   /* out: name array with end of str */
                      cmatrix *Seqchars); /* out: alignment matrix           */

/* read subsets from file */
void readsubsetfile(FILE    *seqfp,              /* in:  sequence input file stream */
                    int      Maxspc,             /* in:  number of taxa             */
                    cmatrix  namestr,            /* in:  names taxa (seq file)      */
                    int     *Maxsubset,          /* out: number of subsets          */
                    imatrix *ss_setovlgraph,     /* out: size of overlap >= 3 between 2 subsets */
                    imatrix *ss_setoverlaps,     /* out: size of overlap between 2 subsets */
                    imatrix *ss_setovllist,      /* out: list with ovlerlapping subsets */
                    ivector *ss_setovllistsize,  /* out: size of list with ovlerlapping subsets */
                    imatrix *ss_matrix,          /* out: boolean list: taxon in set? */
                    imatrix *ss_list,            /* out: list of taxa in set */
                    ivector *ss_listsize);        /* out: size of list with taxa */

/* permute taxon order */
void permutetaxa_ss(int      Maxspc,        /* in:  number of taxa           */
                    ivector  permutation,   /* permuted taxon order          */
                    cmatrix  namestr,       /* in:  names taxa (seq file)    */
                    int      Maxsubset,     /* out: number of subsets        */
                    imatrix  ss_setovlgraph,/* out: size of overlap >= 3 between 2 subsets */
                    imatrix  ss_setoverlaps,/* out: size of overlap between 2 subsets */
                    imatrix  ss_setovllist, /* out: list with overlapping subsets */
                    ivector  ss_setovllistsize,/* out: size of list with ovlerlapping subsets */
                    imatrix  ss_matrix,     /* out: bool list: taxon in set? */
                    imatrix  ss_list,       /* out: list of taxa in set */
                    ivector  ss_listsize);  /* out: size of list with taxa */

/* print subsets */
void fprintfss(FILE    *ofp,              /* in:  output file stream         */
             int      Maxspc,             /* in:  number of taxa             */
             int      Maxsubset,          /* out: number of subsets          */
             imatrix  ss_setovlgraph,     /* out: size of overlap >= 3 between 2 subsets */
             imatrix  ss_setoverlaps,     /* out: size of overlap between 2 subsets */
             imatrix  ss_setovllist,      /* out: list with ovlerlapping subsets */
             ivector  ss_setovllistsize,  /* out: size of list with ovlerlapping subsets */
             imatrix  ss_matrix,          /* out: boolean list: taxon in set? */
             imatrix  ss_list,            /* out: list of taxa in set */
             ivector  ss_listsize);        /* out: size of list with taxa */

/* check connectedness of subsets */
void checkss(FILE    *ofp,              /* in:  output file stream         */
             int      Maxspc,             /* in:  number of taxa             */
             int      Maxsubset,          /* out: number of subsets          */
             imatrix  ss_setovlgraph,     /* out: size of overlap >= 3 between 2 subsets */
             imatrix  ss_setoverlaps,     /* out: size of overlap between 2 subsets */
             imatrix  ss_setovllist,      /* out: list with ovlerlapping subsets */
             ivector  ss_setovllistsize,  /* out: size of list with ovlerlapping subsets */
             imatrix  ss_matrix,          /* out: boolean list: taxon in set? */
             imatrix  ss_list,            /* out: list of taxa in set */
             ivector  ss_listsize);        /* out: size of list with taxa */

/* guess data type: NUCLEOTIDE:0, AMINOACID:1, BINARY:2 */
int guessdatatype(cmatrix Seqchars,   /* alignment matrix (Maxspc x Maxseqc) */
                  int     Maxspc,     /* number of taxa */
                  int     Maxseqc);   /* number of sites */

void translatedataset(int maxspc, int maxseqc, int *maxsite, cmatrix seqchars, cmatrix *seqchar, ivector *seqgapchar, ivector *seqotherchar);
void estimatebasefreqs(void);
void guessmodel(void);





unsigned char *callocquartets(int);
void freequartets(void);
unsigned char readquartet(int, int, int, int);
void writequartet(int, int, int, int, unsigned char);
void sort3doubles(dvector, ivector);
void computeallquartets(void);

/* check the branching structure between the leaves (not the taxa!)
   A, B, C, and I (A, B, C, I don't need to be ordered). As a result,
   the two leaves that are closer related to each other than to leaf I
   are found in chooseX and chooseY. If the branching structure is
   not uniquely defined, chooseX and chooseY are chosen randomly
   from the possible taxa */
void checkquartet(int  A,          /* quartet taxon ID                 */
                  int  B,          /* dito.                            */
                  int  C,          /* dito.                            */
                  int  I,          /* dito., to be inserted            */
                  int *chooseX,    /* (chooseX,chooseY | x,I)          */
                  int *chooseY);   /* chooseX+Y are non-neighbors of I */

void checkquartet_trueID(int  A,   /* quartet leaf idx in permutation array */
                  int  B,          /* dito.                                 */
                  int  C,          /* dito.                                 */
                  int  I,          /* dito., to be inserted                 */
                  int *trueID,     /* permutation array                     */
                  int *chooseX,    /* (chooseX,chooseY | x,I)               */
                  int *chooseY);   /* chooseX+Y are non-neighbors of I      */

void num2quart(uli qnum, int *a, int *b, int *c, int *d);
uli numquarts(int maxspc);
uli quart2num (int a, int b, int c, int d);

void writetpqfheader(int nspec, FILE *ofp, int flag);


/* extracted from main (xxx) */
void compute_quartlklhds(int a, int b, int c, int d, double *d1, double *d2, double *d3, int approx);


/* definitions for timing */

#define OVERALL   0
#define GENERAL   1
#define OPTIONS   2
#define PARAMEST  3
#define QUARTETS  4
#define PUZZLING  5
#define TREEEVAL  6

typedef struct {
	int      currentjob;
	clock_t  tempcpu;
	clock_t  tempfullcpu;
	clock_t  tempcpustart;
	time_t   temptime;
	time_t   tempfulltime;
	time_t   temptimestart;

        clock_t  maxcpu;
	clock_t  mincpu;
	time_t   maxtime;
	time_t   mintime;

	double   maxcpublock;
	double   mincpublock;
	double   mincputick;
	double   mincputicktime;
	double   maxtimeblock;
	double   mintimeblock;

	double   generalcpu;
	double   optionscpu;
	double   paramestcpu;
	double   quartcpu;
	double   quartblockcpu;
	double   quartmaxcpu;
	double   quartmincpu;
	double   puzzcpu;
	double   puzzblockcpu;
	double   puzzmaxcpu;
	double   puzzmincpu;
	double   treecpu;
	double   treeblockcpu;
	double   treemaxcpu;
	double   treemincpu;
	double   cpu;
	double   fullcpu;

	double   generaltime;
	double   optionstime;
	double   paramesttime;
	double   quarttime;
	double   quartblocktime;
	double   quartmaxtime;
	double   quartmintime;
	double   puzztime;
	double   puzzblocktime;
	double   puzzmaxtime;
	double   puzzmintime;
	double   treetime;
	double   treeblocktime;
	double   treemaxtime;
	double   treemintime;
	double   time;
	double   fulltime;
} timearray_t;

 double cputime, walltime;
 double fullcpu, fulltime;
 double fullcputime, fullwalltime;
 double altcputime, altwalltime;
 clock_t cputimestart,  cputimestop, cputimedummy;
 time_t  walltimestart, walltimestop, walltimedummy;
 clock_t Startcpu;     /* start cpu time */
 clock_t Stopcpu;      /* stop cpu time */
 time_t MLstepStarttime;     /* start time */
 time_t MLstepStoptime;      /* stop time */
 time_t PStepStarttime;     /* start time */
 time_t PStepStoptime;      /* stop time */
 time_t PEstStarttime;     /* start time */
 time_t PEstStoptime;      /* stop time */
 time_t Starttime;     /* start time */
 time_t Stoptime;      /* stop time */
 time_t time0;         /* timer variable */
 time_t time1;         /* yet another timer */
 time_t time2;         /* yet another timer */
 timearray_t tarr;

void resetqblocktime(timearray_t *ta);
void resetpblocktime(timearray_t *ta);
void inittimearr(timearray_t *ta);
void addtimes(int jobtype, timearray_t *ta);

/* compute Bayesian weights from log-lkls d1, d2, d3 */
unsigned char loglkl2weight(int    a,
                            int    b,
                            int    c,
                            int    i,
                            double d1,
                            double d2,
                            double d3,
                            int    usebestq);

void memcleanup(void);
void map_lklhd(void);
void recon_tree(void);
void setdefaults(void);
